CREATE TABLE IF NOT EXISTS segment_type (
    segment_type_id smallint NOT NULL,
    segment_tag smallint,
    name VARCHAR(255) NOT NULL,
    platforms VARCHAR(255) NOT NULL,
    packaging smallint NOT NULL,
    version_scheme smallint NOT NULL,
    version_types VARCHAR(255) NOT NULL,
    segment_scopes VARCHAR(255) NOT NULL,
    assign_scopes VARCHAR(255) NOT NULL,
    allow_app boolean DEFAULT FALSE,
    PRIMARY KEY (segment_type_id)
);

CREATE TABLE IF NOT EXISTS segment (
    segment_id bigserial NOT NULL,
    scope_urn VARCHAR(255) NOT NULL,
    name VARCHAR(255) NOT NULL,
    description VARCHAR(255)  NOT NULL,
    segment_type smallint NOT NULL,
    status smallint NOT NULL,
    created_at bigint NOT NULL,
    created_by varchar(255) NOT NULL,
    updated_at bigint  NOT NULL,
    updated_by varchar(255) NOT NULL,
    PRIMARY KEY (segment_id),
    FOREIGN KEY (segment_type) REFERENCES segment_type(segment_type_id)
);

CREATE TABLE IF NOT EXISTS segment_data (
    segment_data_id bigserial NOT NULL,
    segment_id bigint NOT NULL,
    platform smallint NOT NULL,
    major bigint NOT NULL,
    minor bigint NOT NULL,
    body text NOT NULL,
    dialect varchar(255),
    created_at bigint NOT NULL,
    created_by varchar(255) NOT NULL,
    updated_at bigint  NOT NULL,
    updated_by varchar(255) NOT NULL,
    PRIMARY KEY (segment_data_id),
    FOREIGN KEY (segment_id) REFERENCES segment(segment_id) ON DELETE CASCADE,
    UNIQUE (segment_id, platform, major, minor)
);

CREATE TABLE IF NOT EXISTS assignment (
    assignment_id bigserial NOT NULL,
    assignment_scope_urn VARCHAR(255) NOT NULL,
    segment_id bigint NOT NULL,
    platform smallint NOT NULL,
    version_type smallint NOT NULL,
    version_major bigint NOT NULL DEFAULT(-1),
    version_minor bigint NOT NULL DEFAULT(-1),
    status smallint NOT NULL,
    segment_type smallint NOT NULL,
    created_at bigint NOT NULL,
    created_by varchar(255) NOT NULL,
    updated_at bigint  NOT NULL,
    updated_by varchar(255) NOT NULL,
    PRIMARY KEY (assignment_id),
    FOREIGN KEY (segment_id) REFERENCES segment(segment_id) ON DELETE CASCADE,
    FOREIGN KEY (segment_type) REFERENCES segment_type(segment_type_id),
    UNIQUE (assignment_scope_urn, segment_id, platform)
);

CREATE TABLE IF NOT EXISTS app_data (
    app_data_id bigserial NOT NULL,
    segment_id bigint NOT NULL,
    major bigint NOT NULL,
    minor bigint NOT NULL,
    body text NOT NULL,
    created_at bigint NOT NULL,
    created_by varchar(255) NOT NULL,
    updated_at bigint  NOT NULL,
    updated_by varchar(255) NOT NULL,
    PRIMARY KEY (app_data_id),
    FOREIGN KEY (segment_id) REFERENCES segment(segment_id) ON DELETE CASCADE
);

CREATE TABLE IF NOT EXISTS segment_version (
    segment_version_id bigserial NOT NULL,
    segment_data_id bigint NOT NULL,
    segment_id bigint NOT NULL,
    platform smallint NOT NULL,
    version_type smallint NOT NULL,
    version_major bigint NOT NULL DEFAULT(-1),
    version_minor bigint NOT NULL DEFAULT(-1),
    effective_major bigint NOT NULL,
    effective_minor bigint NOT NULL,
    created_at bigint NOT NULL,
    updated_at bigint  NOT NULL,
    PRIMARY KEY (segment_version_id),
    FOREIGN KEY (segment_id) REFERENCES segment(segment_id) ON DELETE CASCADE,
    FOREIGN KEY (segment_data_id) REFERENCES segment_data(segment_data_id) ON DELETE CASCADE
);

CREATE TABLE IF NOT EXISTS ruleset (
    ruleset_id varchar(255) NOT NULL,
    signature_hash varchar(255) NOT NULL,
    data text NOT NULL,
    created_at bigint NOT NULL,
    PRIMARY KEY (ruleset_id)
);

CREATE TABLE IF NOT EXISTS ruleset_data (
    ruleset_id varchar(255) NOT NULL,
    segment_data_id bigint NOT NULL,
    PRIMARY KEY (ruleset_id, segment_data_id),
    FOREIGN KEY (ruleset_id) REFERENCES ruleset(ruleset_id) ON DELETE CASCADE,
    FOREIGN KEY (segment_data_id) REFERENCES segment_data(segment_data_id)
);

-- Matches WHERE clause in /meta call with COVERING fields
CREATE INDEX IF NOT EXISTS idx_assignment_version_join ON assignment (
  assignment_scope_urn,
  platform,
  status,
  segment_id,
  version_type,
  version_major NULLS LAST,
  version_minor NULLS FIRST,
  updated_at,
  segment_type
)
WHERE
  status = 1;

-- Matches WHERE clause in /meta call with COVERING fields
CREATE INDEX IF NOT EXISTS idx_segment_version_join ON segment_version (
  segment_id,
  platform,
  version_type,
  version_major NULLS LAST,
  version_minor NULLS FIRST,
  segment_data_id,
  effective_major,
  effective_minor,
  updated_at
);

DO
$do$
BEGIN
IF NOT EXISTS(select tgname from pg_trigger where tgname='trigger_segment_versions') THEN
    RAISE NOTICE 'Creating segment data trigger';
    CREATE OR REPLACE FUNCTION process_segment_versions() RETURNS TRIGGER AS $trigger_segment_versions$
    DECLARE
        PINNED          smallint := 0;
        NON_BREAKING    smallint := 1;
        LATEST          smallint := 2;
        NA              smallint := -1;
        EXISTING_NON_BREAKING_WITH_MAJOR segment_version%ROWTYPE;
        EXISTING_LATEST segment_version%ROWTYPE;
        SEGMENT_DATA_EXISTS boolean := (select exists
            (SELECT 1 FROM segment_data WHERE segment_data.segment_id = NEW.segment_id));
    BEGIN
        --
        -- Trigger to insert or update segment_version records based on new segment_data
        --
        IF (TG_OP = 'INSERT') THEN
            -- ALWAYS insert a new PINNED segment_version
            INSERT INTO segment_version values (
                    DEFAULT, NEW.segment_data_id, NEW.segment_id, NEW.platform, PINNED,
                    NEW.major, NEW.minor, NEW.major, NEW.minor,
                    NEW.created_at,  NEW.updated_at);

            IF NOT SEGMENT_DATA_EXISTS THEN
                -- NON_BREAKING
                INSERT INTO segment_version values (
                        DEFAULT, NEW.segment_data_id, NEW.segment_id, NEW.platform,
                        NON_BREAKING, NEW.major, NA, NEW.major, NEW.minor,
                        NEW.created_at,  NEW.updated_at);
                -- LATEST
                INSERT INTO segment_version values (
                            DEFAULT, NEW.segment_data_id, NEW.segment_id, NEW.platform,
                            LATEST, NA, NA, NEW.major, NEW.minor,
                            NEW.created_at,  NEW.updated_at);
            ELSE
                SELECT * INTO EXISTING_NON_BREAKING_WITH_MAJOR FROM segment_version
                    WHERE
                        segment_id = NEW.segment_id
                        AND version_type = NON_BREAKING
                        AND effective_major = NEW.major
                        AND platform = NEW.platform;

                IF(EXISTING_NON_BREAKING_WITH_MAJOR IS NULL) THEN
                    -- No NON_BREAKING ENTRY with this major version
                    INSERT INTO segment_version values (
                            DEFAULT, NEW.segment_data_id, NEW.segment_id, NEW.platform,
                            NON_BREAKING, NEW.major, NA, NEW.major, NEW.minor,
                            NEW.created_at,  NEW.updated_at);
                ELSEIF (EXISTING_NON_BREAKING_WITH_MAJOR.effective_minor < NEW.minor) THEN
                    -- Update NON_BREAKING
                    UPDATE segment_version
                        SET
                            segment_data_id = NEW.segment_data_id,
                            version_major = NEW.major,
                            effective_major = NEW.major,
                            effective_minor = NEW.minor,
                            updated_at = NEW.updated_at
                        WHERE
                            segment_version_id = EXISTING_NON_BREAKING_WITH_MAJOR.segment_version_id;
                END IF;

                SELECT * INTO EXISTING_LATEST FROM segment_version
                    WHERE
                        segment_id = NEW.segment_id
                        AND version_type = LATEST
                        AND platform = NEW.platform;

                IF(EXISTING_LATEST IS NULL) THEN
                    -- No LATEST entry at all
                    INSERT INTO segment_version values (
                            DEFAULT, NEW.segment_data_id, NEW.segment_id, NEW.platform,
                            LATEST, NA, NA, NEW.major, NEW.minor,
                            NEW.created_at,  NEW.updated_at);
                ELSEIF (
                    (EXISTING_LATEST.effective_major = NEW.major AND EXISTING_LATEST.effective_minor < NEW.minor)
                        OR EXISTING_LATEST.effective_major < NEW.major) THEN
                    -- Update LATEST
                    UPDATE segment_version
                        SET
                            segment_data_id = NEW.segment_data_id,
                            effective_major = NEW.major,
                            effective_minor = NEW.minor,
                            updated_at = NEW.updated_at
                        WHERE
                            segment_version_id = EXISTING_LATEST.segment_version_id;
                END IF;
            END IF;
        END IF;
        RETURN NULL; -- result is ignored since this is an AFTER trigger
    END;
    $trigger_segment_versions$ LANGUAGE plpgsql;

    CREATE TRIGGER trigger_segment_versions
        AFTER INSERT ON segment_data
        FOR EACH ROW EXECUTE PROCEDURE process_segment_versions();
ELSE
    RAISE NOTICE 'Segment Version Trigger already created.';
END IF;
END
$do$;